# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Timekeeping
-----------

The ``drgn.helpers.linux.timekeeping`` module provides helpers for timestamps.

.. note::

    In core dumps, it is only possible to recover coarse timestamps, which are
    only updated once per tick (~1-10 ms). Therefore, only the
    ``ktime_get_seconds()`` and ``ktime_get_coarse_ns()`` families of functions
    are provided. These functions can also race with timekeeping updates and
    return a value with an error of up to 1 second.
"""

import functools
import logging
import re
from typing import Optional

from drgn import Object, ObjectNotFoundError, Program, ProgramFlags, cast, sizeof
from drgn.helpers.common.prog import takes_program_or_default

logger = logging.getLogger("drgn")

__all__ = (
    "ktime_get_boottime_seconds",
    "ktime_get_clocktai_seconds",
    "ktime_get_coarse_boottime_ns",
    "ktime_get_coarse_clocktai_ns",
    "ktime_get_coarse_ns",
    "ktime_get_coarse_real_ns",
    "ktime_get_real_seconds",
    "ktime_get_seconds",
    "ktime_to_ns",
    "uptime",
    "uptime_pretty",
)


def ktime_to_ns(kt: Object) -> Object:
    """
    Convert a ``ktime_t`` to nanoseconds.

    :param kt: ``ktime_t``
    :return: ``s64``
    """
    # Since Linux kernel commit 2456e8553544 ("ktime: Get rid of the union")
    # (in v4.10), ktime_t is a typedef of s64. Before that, it was a dummy
    # union wrapping an s64.
    try:
        return cast("s64", kt)
    except TypeError:
        return kt.tv64.read_()


def _tk_core(prog: Program) -> Object:
    try:
        return prog.cache["tk_core"]
    except KeyError:
        pass
    # Since Linux kernel commit 22c62b9a84b8 ("timekeeping: Introduce auxiliary
    # timekeepers") (in v6.17), tk_core is a macro for an array element.
    try:
        tk_core = prog["timekeeper_data"][prog["TIMEKEEPER_CORE"]]
    except ObjectNotFoundError:
        tk_core = prog["tk_core"]
    prog.cache["tk_core"] = tk_core
    return tk_core


@takes_program_or_default
def ktime_get_seconds(prog: Program) -> Object:
    """
    Get the seconds component of the monotonic time
    (``CLOCK_MONOTONIC_COARSE``).

    :return: ``time64_t``
    """
    return cast("time64_t", _tk_core(prog).timekeeper.ktime_sec)


def _ktime_get_coarse_ns(prog: Program, offs_name: Optional[str]) -> Object:
    # The timekeeper is updated every tick, so we're very likely to catch it in
    # the middle of an update. To make that less likely, we take a "snapshot"
    # of the relevant timekeeper memory, check whether the seqcount was
    # write-locked during our snapshot, and retry (if live) or warn if so. The
    # snapshot isn't atomic, so it's still possible to race, but the window is
    # narrow.
    #
    # An alternative that fully avoids the race would be simulating a
    # read_seqlock operation, but that requires multiple reads and would likely
    # never succeed on, for example, slow remote connections.
    #
    # We cache the address and size to read and functions for extracting
    # members.
    try:
        address, size, from_snapshot = prog.cache["timekeeper_snapshot"]
    except KeyError:
        tk_core = _tk_core(prog)
        tk = tk_core.timekeeper
        members = {
            "base": tk.tkr_mono.base,
            "offs_real": tk.offs_real,
            "offs_boot": tk.offs_boot,
            "offs_tai": tk.offs_tai,
        }

        # Since Linux kernel commit 025e82bcbc34 ("timekeeping: Use sequence
        # counter with associated raw spinlock") (in v5.9), struct tk_data::seq
        # is a seqcount_raw_spinlock_t. Before that, it is a seqcount_t.
        try:
            members["seq"] = tk_core.seq.seqcount.sequence
        except AttributeError:
            members["seq"] = tk_core.seq.sequence

        # Since Linux kernel commit b71f9804f66c ("timekeeping: Prevent coarse
        # clocks going backwards") (in v6.15), the nanoseconds part is directly
        # in struct timekeeper::coarse_nsec. Before that, it is computed from
        # struct timekeeper::tkr_mono.
        try:
            members["coarse_nsec"] = tk.coarse_nsec
        except AttributeError:
            members["xtime_nsec"] = tk.tkr_mono.xtime_nsec
            members["shift"] = tk.tkr_mono.shift

        address = min(  # type: ignore[type-var]  # member.address_ can't be None
            member.address_ for member in members.values()
        )
        size = (
            max(
                member.address_ + sizeof(member)  # type: ignore[operator]  # member.address_ can't be None.
                for member in members.values()
            )
            - address
        )
        assert size <= 1024

        from_snapshot = {}
        for name, member in members.items():
            member_type = member.type_
            if member_type.type_name() == "ktime_t":
                # Bypass ktime_to_ns() and interpret ktime_t as u64.
                member_type = prog.type("u64")
            from_snapshot[name] = functools.partial(
                Object.from_bytes_,
                prog,
                member_type,
                bit_offset=(member.address_ - address) * 8,
            )

        prog.cache["timekeeper_snapshot"] = address, size, from_snapshot

    # On live kernels, we retry a limited number of times, then warn and move
    # on.
    for _ in range(1000):
        snapshot = prog.read(address, size)
        seq = from_snapshot["seq"](snapshot)
        if seq & 1:
            if prog.flags & ProgramFlags.IS_LIVE:
                continue
            else:
                # For core dumps, the best we can do is warn.
                logger.warning("timekeeper was write-locked; ktime may be inconsistent")
                break
        break
    else:
        logger.warning(
            "couldn't get unlocked snapshot of timekeeper; ktime may be inconsistent"
        )

    ns = from_snapshot["base"](snapshot)
    try:
        ns += from_snapshot["coarse_nsec"](snapshot)
    except KeyError:
        ns += from_snapshot["xtime_nsec"](snapshot) >> from_snapshot["shift"](snapshot)
    if offs_name is not None:
        ns += from_snapshot[offs_name](snapshot)
    return ns


@takes_program_or_default
def ktime_get_coarse_ns(prog: Program) -> Object:
    """
    Get the coarse monotonic time in nanoseconds (``CLOCK_MONOTONIC_COARSE``).

    :return: ``u64``
    """
    return _ktime_get_coarse_ns(prog, None)


@takes_program_or_default
def ktime_get_real_seconds(prog: Program) -> Object:
    """
    Get the seconds component of the real (wall) time
    (``CLOCK_REALTIME_COARSE``).

    :return: ``time64_t``
    """
    match = re.search(
        b"^CRASHTIME=([0-9]+)$", prog["VMCOREINFO"].string_(), flags=re.MULTILINE
    )
    if match:
        try:
            time_type = prog.type("time64_t")
        except LookupError:
            # We want this to work even without any debugging information, so
            # fall back to long long.
            time_type = prog.type("long long")
        return Object(prog, time_type, int(match.group(1)))
    return cast("time64_t", _tk_core(prog).timekeeper.xtime_sec)


@takes_program_or_default
def ktime_get_coarse_real_ns(prog: Program) -> Object:
    """
    Get the coarse real (wall) time in nanoseconds (``CLOCK_REALTIME_COARSE``).

    :return: ``u64``
    """
    return _ktime_get_coarse_ns(prog, "offs_real")


@takes_program_or_default
def ktime_get_boottime_seconds(prog: Program) -> Object:
    """
    Get the seconds component of the monotonic time since boot (coarse version
    of ``CLOCK_BOOTTIME``).

    :return: ``time64_t``
    """
    return cast("time64_t", ktime_get_coarse_boottime_ns(prog) / 1000000000)


@takes_program_or_default
def ktime_get_coarse_boottime_ns(prog: Program) -> Object:
    """
    Get the the coarse monotonic time since boot in nanoseconds (coarse version
    of ``CLOCK_BOOTTIME``).

    :return: ``u64``
    """
    return _ktime_get_coarse_ns(prog, "offs_boot")


@takes_program_or_default
def ktime_get_clocktai_seconds(prog: Program) -> Object:
    """
    Get the seconds component of the International Atomic Time (coarse version
    of ``CLOCK_TAI``).

    :return: ``time64_t``
    """
    return cast("time64_t", ktime_get_coarse_clocktai_ns(prog) / 1000000000)


@takes_program_or_default
def ktime_get_coarse_clocktai_ns(prog: Program) -> Object:
    """
    Get the coarse International Atomic Time in nanoseconds (coarse version of ``CLOCK_TAI``).

    :return: ``u64``
    """
    return _ktime_get_coarse_ns(prog, "offs_tai")


@takes_program_or_default
def uptime(prog: Program) -> float:
    """Get the system uptime (as of the last tick) in fractional seconds."""
    return ktime_get_coarse_boottime_ns(prog).value_() / 1e9


@takes_program_or_default
def uptime_pretty(prog: Program) -> str:
    """
    Get the system uptime as a human-readable string.

    >>> uptime_pretty()
    '1 day, 6 hours, 56 minutes, 40 seconds'
    """
    seconds = ktime_get_boottime_seconds(prog).value_()

    parts = []
    for unit, seconds_in_unit in (
        ("year", 365 * 24 * 60 * 60),
        ("week", 7 * 24 * 60 * 60),
        ("day", 24 * 60 * 60),
        ("hour", 60 * 60),
        ("minute", 60),
    ):
        if seconds >= seconds_in_unit:
            units = seconds // seconds_in_unit
            seconds -= units * seconds_in_unit
            s = "" if units == 1 else "s"
            parts.append(f"{units} {unit}{s}")
    if seconds or not parts:
        s = "" if seconds == 1 else "s"
        parts.append(f"{seconds} second{s}")
    return ", ".join(parts)



# Copyright (c) 2024, Oracle and/or its affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""Compatibility shim between drgn and the pyrepl/code modules"""

import builtins
import os
from typing import TYPE_CHECKING, Any, Dict

from drgn.commands import _write_command_error, run_command

__all__ = ("interact", "readline")

# Python 3.13 introduces a new REPL implemented by the "_pyrepl" internal
# module. It includes features such as colored output and multiline editing.
# Unfortunately, there is no public API exposing these abilities to users, even
# in the "code" module. We'd like to give the best experience possible, so we'll
# detect _pyrepl and try to use it where possible.
try:
    # The official Python interpreter honors this environment variable to
    # disable the new REPL. We do the same, which also gives users an escape
    # hatch if any of the internals we're messing with change.
    if os.environ.get("PYTHON_BASIC_REPL"):
        raise ModuleNotFoundError()

    # Unfortunately, the typeshed library behind mypy explicitly removed type
    # stubs for these modules. This makes sense as they are private APIs, but it
    # means we need to disable mypy checks.
    from _pyrepl import readline  # type: ignore
    from _pyrepl.simple_interact import (  # type: ignore
        run_multiline_interactive_console,
    )

    # Do this instead of type: ignore so that _BaseConsole gets type-checked as
    # code.InteractiveConsole instead of Any.
    if not TYPE_CHECKING:
        from _pyrepl.console import InteractiveColoredConsole as _BaseConsole

    # This _setup() function clobbers the readline completer, but it is
    # protected so it only runs once. Call it early so that our overridden
    # completer doesn't get clobbered.
    readline._setup({})

    def interact(local: Dict[str, Any], banner: str) -> None:
        console = _InteractiveConsoleWithCommand(local)
        console.write(banner + "\n")
        run_multiline_interactive_console(console)

except (ModuleNotFoundError, ImportError):
    from code import InteractiveConsole as _BaseConsole
    import readline

    def interact(local: Dict[str, Any], banner: str) -> None:
        _InteractiveConsoleWithCommand(local).interact(banner, exitmsg="")


class _InteractiveConsoleWithCommand(_BaseConsole):
    def _on_command_error(self, e: Exception) -> None:
        return _write_command_error(self, e)

    def runsource(
        self, source: str, filename: str = "<input>", symbol: str = "single"
    ) -> bool:
        if not source or source[0] != "%":
            return super().runsource(source, filename, symbol)

        ret = run_command(
            self.locals["prog"],
            source[1:],
            # The version of typeshed that we're stuck on says locals is a
            # Mapping[str, Any], but upstream corrected it to dict[str, Any].
            globals=self.locals,  # type: ignore[arg-type]
            onerror=self._on_command_error,
        )
        if ret is not None:
            builtins._ = ret  # type: ignore[attr-defined]
        # In theory, we could allow splitting pipelines over multiple lines
        # like sh, but run_multiline_interactive_console() doesn't support
        # console.runsource() returning True for more lines.
        return False



#!/usr/bin/env drgn

# Copyright (c) 2025 NVIDIA Corporation & Affiliates
# SPDX-License-Identifier: LGPL-2.1-or-later

help="""
Takes in a virtual address from a page or a `struct page *` and tries to find
a SKB that references this page either in the linear part or in skb_frag_info.

Endianess is handled by the script.

This only works if the kernel was built with CONFIG_PROC_KCORE=y. The
script is based on search_kernel_memory.py
"""

import argparse
import math
import sys
from drgn import (
        Object,
        PlatformFlags,
        FaultError,
        Object,
        offsetof,
        sizeof,
)
from drgn.helpers.common.memory import (
        identify_address,
)
from drgn.helpers.linux.list import list_for_each_entry
from drgn.helpers.linux.mm import (
        for_each_vmap_area,
        virt_to_page,
        page_to_virt,
)
from drgn.helpers.linux.net import (
    skb_shinfo,
)


byteorder = "little" if prog.platform.flags & PlatformFlags.IS_LITTLE_ENDIAN else "big"
PAGE_SIZE = prog["PAGE_SIZE"].value_()
PAGE_SHIFT = prog["PAGE_SHIFT"].value_()


def get_opts():
    parser = argparse.ArgumentParser(description=help)
    parser.add_argument(
        "bytes",
        nargs="?",
        help="hexadecimal bytes to read. By default they represent a "
             "virtual address.",
    )
    parser.add_argument(
        "--as-frag", default=False, action="store_true",
        help="Interpret address as being skb_shinfo(skb).frag.netmem.")
    parser.add_argument(
        "--virt", default=False, action="store_true",
        help="Given address is a virtual addresss in a page.")
    parser.add_argument(
        "--show-skb", default=False, action="store_true",
        help="Show matched SKB.")
    parser.add_argument(
        "--verbose", default=False, action="store_true",
        help="Print partial matches.")

    return parser.parse_args()


def virt_to_vmap_address(prog, addr):
    page = virt_to_page(addr)
    for va in for_each_vmap_area(prog):
        vm = va.vm.read_()
        if vm:
            for i, va_page in enumerate(
                Object(
                    prog, prog.array_type(page.type_, vm.nr_pages), address=vm.pages
                ).read_()
            ):
                if va_page == page:
                    return (
                        va.va_start.value_()
                        + (i << prog["PAGE_SHIFT"])
                        + (addr & (prog["PAGE_SIZE"].value_() - 1))
                    )
    return None


def search_memory(prog, needle):
    KCORE_RAM = prog["KCORE_RAM"]
    CHUNK_SIZE = 1024 * 1024

    for kc in list_for_each_entry(
        "struct kcore_list", prog["kclist_head"].address_of_(), "list"
    ):
        if kc.type != KCORE_RAM:
            continue
        start = kc.addr.value_()
        end = start + kc.size.value_()
        for addr in range(start, end, CHUNK_SIZE):
            buf = prog.read(addr, min(CHUNK_SIZE, end - addr))
            i = 0
            while i < len(buf):
                i = buf.find(needle, i)
                if i < 0:
                    break

                yield addr + i
                i += 8


def search_page_reference(page):
    """
    Search kernel memory for references to the given page contents
    (virtual addresses within the PAGE_SIZE range).

    Does page conversion.
    """

    val = page_to_virt(page).value_()

    skip_bytes = math.ceil(PAGE_SHIFT / 8)
    ptr_size = 8

    val_endian = val.to_bytes(ptr_size, byteorder)
    if byteorder == "little":
        big_needle = val_endian[skip_bytes:ptr_size - skip_bytes]
    else:
        big_needle = val_endian[0:ptr_size - skip_bytes]

    small_needle = val >> PAGE_SHIFT

    # Search for first 6 bytes:
    for addr in search_memory(prog, big_needle):

        if byteorder == "little":
            # Adjust address to skipped bytes:
            addr = addr - skip_bytes

        mem_bytes = prog.read(addr, ptr_size)
        mem_val = int.from_bytes(mem_bytes, byteorder)

        if mem_val >> PAGE_SHIFT == small_needle:
            yield (addr, mem_val)


def search_raw(value):
    """
    Search kernel memory for value respectinv the value pointer.
    """

    ptr_size = 8
    needle = value.to_bytes(ptr_size, byteorder)

    for addr in search_memory(prog, needle):
        mem_bytes = prog.read(addr, ptr_size)
        mem_val = int.from_bytes(mem_bytes, byteorder)
        yield (addr, mem_val)


def guess_skb_is_legit(skb) -> bool:
    """
    Guess if there is a legit SKB at the given address.
    """

    # 2 consecutive pointers that point to the same page indicate that
    # this could be skb.head and skb.data.
    if virt_to_page(skb.head).value_() != virt_to_page(skb.data).value_():
        return False

    if skb.end.value_() > PAGE_SIZE or skb.tail.value_() > PAGE_SIZE:
        return False

    # Many checks could be added here ...
    return True


def search_skb_with_page_as_linear(page, verbose=False):
    """
    Search SKB for given page.
    """
    for addr, val in search_page_reference(page):

        if verbose:
            print(f"Found reference at {hex(addr)}: value {hex(val)}. {identify_address(prog, addr)}")

        skb_addr = addr - offsetof(prog.type("struct sk_buff"), "head")
        skb = Object(prog, "struct sk_buff", address=skb_addr)
        if guess_skb_is_legit(skb):
            yield skb


def search_skb_with_page_as_shinfo_frag(page_ptr, verbose):

    for addr, _ in search_raw(page_ptr):

        if verbose:
            print(f"Found raw value at addr {hex(addr)}. {identify_address(prog, addr)}")

        page = virt_to_page(addr)
        for skb in search_skb_with_page_as_linear(page):

            # For shinfo, a match happens for
            shinfo = skb_shinfo(skb)
            shinfo_start = shinfo.value_()
            shinfo_end = shinfo.value_() + sizeof(prog.type("struct skb_shared_info"))
            if shinfo_start <= addr and addr < shinfo_end:
                yield skb


opts = get_opts()

# Drop hex prefix.
if opts.bytes.startswith("0x"):
    opts.bytes = opts.bytes[2:]

value = int.from_bytes(bytes.fromhex(opts.bytes))

if opts.as_frag:
    for skb in search_skb_with_page_as_shinfo_frag(value, opts.verbose):
        print(f"Possible skb match at address {hex(skb.address_of_())}")
        if opts.show_skb:
            print(skb)

else:
    if opts.virt:
        try:
            page = virt_to_page(value)
        except FaultError:
            print("Given address doesn't seem to be a virtual address or it can't be converted to a page.")
            sys.exit(1)
    else:
        page = Object(prog, "struct page", address=value).address_of_()

    for skb in search_skb_with_page_as_linear(page, opts.verbose):
        print(f"Possible skb match at address {hex(skb.address_of_())}")
        if opts.show_skb:
            print(skb)




# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Linked Lists
------------

The ``drgn.helpers.linux.list`` module provides helpers for working with the
doubly-linked list implementations (``struct list_head`` and ``struct
hlist_head``) in :linux:`include/linux/list.h`.
"""

from typing import Iterator, Union

from drgn import NULL, Object, Type, container_of
from drgn.helpers import ValidationError

__all__ = (
    "hlist_empty",
    "hlist_for_each",
    "hlist_for_each_entry",
    "list_count_nodes",
    "list_empty",
    "list_first_entry",
    "list_first_entry_or_null",
    "list_for_each",
    "list_for_each_entry",
    "list_for_each_entry_reverse",
    "list_for_each_reverse",
    "list_is_singular",
    "list_last_entry",
    "list_next_entry",
    "list_prev_entry",
    "validate_list",
    "validate_list_count_nodes",
    "validate_list_for_each",
    "validate_list_for_each_entry",
)


def list_empty(head: Object) -> bool:
    """
    Return whether a list is empty.

    :param head: ``struct list_head *``
    """
    head = head.read_()
    return head.next == head


def list_is_singular(head: Object) -> bool:
    """
    Return whether a list has only one element.

    :param head: ``struct list_head *``
    """
    head = head.read_()
    next = head.next
    return next != head and next == head.prev


def list_count_nodes(head: Object) -> int:
    """
    Return the number of nodes in a list.

    :param head: ``struct list_head *``
    """
    return sum(1 for _ in list_for_each(head))


def list_first_entry(head: Object, type: Union[str, Type], member: str) -> Object:
    """
    Return the first entry in a list.

    The list is assumed to be non-empty.

    See also :func:`list_first_entry_or_null()`.

    :param head: ``struct list_head *``
    :param type: Entry type.
    :param member: Name of list node member in entry type.
    :return: ``type *``
    """
    return container_of(head.next, type, member)


def list_first_entry_or_null(
    head: Object, type: Union[str, Type], member: str
) -> Object:
    """
    Return the first entry in a list or ``NULL`` if the list is empty.

    See also :func:`list_first_entry()`.

    :param head: ``struct list_head *``
    :param type: Entry type.
    :param member: Name of list node member in entry type.
    :return: ``type *``
    """
    head = head.read_()
    pos = head.next.read_()
    if pos == head:
        return NULL(head.prog_, head.prog_.pointer_type(head.prog_.type(type)))
    else:
        return container_of(pos, type, member)


def list_last_entry(head: Object, type: Union[str, Type], member: str) -> Object:
    """
    Return the last entry in a list.

    The list is assumed to be non-empty.

    :param head: ``struct list_head *``
    :param type: Entry type.
    :param member: Name of list node member in entry type.
    :return: ``type *``
    """
    return container_of(head.prev, type, member)


def list_next_entry(pos: Object, member: str) -> Object:
    """
    Return the next entry in a list.

    :param pos: ``type *``
    :param member: Name of list node member in entry type.
    :return: ``type *``
    """
    return container_of(pos.member_(member).next, pos.type_.type, member)


def list_prev_entry(pos: Object, member: str) -> Object:
    """
    Return the previous entry in a list.

    :param pos: ``type *``
    :param member: Name of list node member in entry type.
    :return: ``type *``
    """
    return container_of(pos.member_(member).prev, pos.type_.type, member)


def list_for_each(head: Object) -> Iterator[Object]:
    """
    Iterate over all of the nodes in a list.

    :param head: ``struct list_head *``
    :return: Iterator of ``struct list_head *`` objects.
    """
    head = head.read_()
    pos = head.next.read_()
    while pos != head:
        yield pos
        pos = pos.next.read_()


def list_for_each_reverse(head: Object) -> Iterator[Object]:
    """
    Iterate over all of the nodes in a list in reverse order.

    :param head: ``struct list_head *``
    :return: Iterator of ``struct list_head *`` objects.
    """
    head = head.read_()
    pos = head.prev.read_()
    while pos != head:
        yield pos
        pos = pos.prev.read_()


def list_for_each_entry(
    type: Union[str, Type], head: Object, member: str
) -> Iterator[Object]:
    """
    Iterate over all of the entries in a list.

    :param type: Entry type.
    :param head: ``struct list_head *``
    :param member: Name of list node member in entry type.
    :return: Iterator of ``type *`` objects.
    """
    type = head.prog_.type(type)
    for pos in list_for_each(head):
        yield container_of(pos, type, member)


def list_for_each_entry_reverse(
    type: Union[str, Type], head: Object, member: str
) -> Iterator[Object]:
    """
    Iterate over all of the entries in a list in reverse order.

    :param type: Entry type.
    :param head: ``struct list_head *``
    :param member: Name of list node member in entry type.
    :return: Iterator of ``type *`` objects.
    """
    type = head.prog_.type(type)
    for pos in list_for_each_reverse(head):
        yield container_of(pos, type, member)


def validate_list(head: Object) -> None:
    """
    Validate that the ``next`` and ``prev`` pointers in a list are consistent.

    >>> validate_list(prog["my_list"].address_of_())
    drgn.helpers.ValidationError: (struct list_head *)0xffffffffc029e460 next 0xffffffffc029e000 has prev 0xffffffffc029e450

    :param head: ``struct list_head *``
    :raises ValidationError: if the list is invalid
    """
    for _ in validate_list_for_each(head):
        pass


def validate_list_count_nodes(head: Object) -> int:
    """
    Like :func:`list_count_nodes()`, but validates the list like
    :func:`validate_list()` while iterating.

    :param head: ``struct list_head *``
    """
    return sum(1 for _ in validate_list_for_each(head))


def validate_list_for_each(head: Object) -> Iterator[Object]:
    """
    Like :func:`list_for_each()`, but validates the list like
    :func:`validate_list()` while iterating.

    :param head: ``struct list_head *``
    :raises ValidationError: if the list is invalid
    """
    head = head.read_()
    pos = head.next.read_()
    while pos != head:
        yield pos
        next = pos.next.read_()
        next_prev = next.prev.read_()
        if next_prev != pos:
            raise ValidationError(
                f"{pos.format_(dereference=False, symbolize=False)}"
                f" next {next.format_(dereference=False, symbolize=False, type_name=False)}"
                f" has prev {next_prev.format_(dereference=False, symbolize=False, type_name=False)}"
            )
        pos = next


def validate_list_for_each_entry(
    type: Union[str, Type], head: Object, member: str
) -> Iterator[Object]:
    """
    Like :func:`list_for_each_entry()`, but validates the list like
    :func:`validate_list()` while iterating.

    .. code-block:: python3

       def validate_my_list(prog):
            for entry in validate_list_for_each_entry(
                "struct my_entry",
                prog["my_list"].address_of_(),
                "list",
            ):
                if entry.value < 0:
                    raise ValidationError("list contains negative entry")

    :param type: Entry type.
    :param head: ``struct list_head *``
    :param member: Name of list node member in entry type.
    :raises ValidationError: if the list is invalid
    """
    type = head.prog_.type(type)
    for pos in validate_list_for_each(head):
        yield container_of(pos, type, member)


def hlist_empty(head: Object) -> bool:
    """
    Return whether a hash list is empty.

    :param head: ``struct hlist_head *``
    """
    return not head.first


def hlist_for_each(head: Object) -> Iterator[Object]:
    """
    Iterate over all of the nodes in a hash list.

    :param head: ``struct hlist_head *``
    :return: Iterator of ``struct hlist_node *`` objects.
    """
    pos = head.first.read_()
    while pos:
        yield pos
        pos = pos.next.read_()


def hlist_for_each_entry(
    type: Union[str, Type], head: Object, member: str
) -> Iterator[Object]:
    """
    Iterate over all of the entries in a hash list.

    :param type: Entry type.
    :param head: ``struct hlist_head *``
    :param member: Name of list node member in entry type.
    :return: Iterator of ``type *`` objects.
    """
    type = head.prog_.type(type)
    for pos in hlist_for_each(head):
        yield container_of(pos, type, member)



#!/usr/bin/env drgn
# Copyright (c) Western Digital Corporation, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""  Dump all block group caches for a given btrfs file-system """

import sys
import drgn
from enum import Flag
from drgn import NULL, Object, cast, container_of, execscript, \
        reinterpret, sizeof
from drgn.helpers.linux import *
from drgn.helpers.common import decode_flags

BTRFS_BLOCK_GROUP_FLAGS = [
        ("BTRFS_BLOCK_GROUP_DATA", 0),
        ("BTRFS_BLOCK_GROUP_SYSTEM", 1),
        ("BTRFS_BLOCK_GROUP_METADATA", 2),
        ("BTRFS_BLOCK_GROUP_RAID0", 3),
        ("BTRFS_BLOCK_GROUP_RAID1", 4),
        ("BTRFS_BLOCK_GROUP_DUP", 5),
        ("BTRFS_BLOCK_GROUP_RAID10", 6),
        ("BTRFS_BLOCK_GROUP_RAID5", 9),
        ("BTRFS_BLOCK_GROUP_RAID6", 8),
        ("BTRFS_BLOCK_GROUP_RAID1C3", 9),
        ("BTRFS_BLOCK_GROUP_RAID1C4", 10)
        ]

BTRFS_BLOCK_GROUP_RUNTIME_FLAGS = [
        ("BLOCK_GROUP_FLAG_IREF", 0),
        ("BLOCK_GROUP_FLAG_REMOVED", 1),
        ("BLOCK_GROUP_FLAG_TO_COPY", 2),
        ("BLOCK_GROUP_FLAG_RELOCATING_REPAIR", 3),
        ("BLOCK_GROUP_FLAG_CHUNK_ITEM_INSERTED", 4),
        ("BLOCK_GROUP_FLAG_ZONE_IS_ACTIVE", 5),
        ("BLOCK_GROUP_FLAG_ZONED_DATA_RELOC", 6),
        ("BLOCK_GROUP_FLAG_NEEDS_FREE_SPACE", 7),
        ("BLOCK_GROUP_FLAG_SEQUENTIAL_ZONE", 8)
        ]

if len(sys.argv) > 1:
    mnt_path = sys.argv[1]
    mnt_path = mnt_path.rstrip('/')
else:
    mnt_path = "/"

mnt = None

for mnt in for_each_mount(prog, dst = mnt_path):
    pass

if mnt is None:
    sys.stderr.write(f'Error: mount point {mnt_path} not found')
    sys.exit(1)

try:
    fs_info = cast('struct btrfs_fs_info *', mnt.mnt.mnt_sb.s_fs_info)
except LookupError:
    print('cannot find \'struct btrfs_fs_info *\', module not loaded?')
    sys.exit(1)

def dump_bg(bg):
    print(f'BG at {bg.start.value_()}')
    print(f'\tflags: {decode_flags(bg.flags.value_(), BTRFS_BLOCK_GROUP_FLAGS)} ({hex(bg.flags)})')
    print(f'\tlength: {bg.length.value_()}')
    print(f'\tused: {bg.used.value_()}')
    print(f'\tpinned: {bg.pinned.value_()}')
    print(f'\treserved: {bg.reserved.value_()}')
    print(f'\truntime_flags: {decode_flags(bg.runtime_flags.value_(), BTRFS_BLOCK_GROUP_RUNTIME_FLAGS)} ({hex(bg.runtime_flags)})')
    if bg.fs_info.zone_size.value_() > 0:
        print(f'\tzone_unsuable: {bg.zone_unusable.value_()}')
    print()

for bg in rbtree_inorder_for_each_entry("struct btrfs_block_group",\
        fs_info.block_group_cache_tree.rb_root, "cache_node"):
    dump_bg(bg)
